﻿using System;
using System.Diagnostics;
using System.Threading;
using System.Runtime.InteropServices;
using System.Collections.Generic;
using System.Linq;
using System.Windows.Forms;

namespace NehebkauWindowsHost
{
    static class Program
    {
        #region Window Management
        [DllImport("user32.dll")]
        private static extern int FindWindow(string className, string windowText);
        [DllImport("user32.dll")]
        private static extern int ShowWindow(int hwnd, int command);
        private const int SW_HIDE = 0;
        private const int SW_SHOW = 1;
        private static bool taskbarIsShown = true;

        public static void HideTaskBar()
        {
            int hwnd = FindWindow("Shell_TrayWnd", "");
            ShowWindow(hwnd, SW_HIDE);
            taskbarIsShown = false;
        }

        public static void ShowTaskBar()
        {
            int hwnd = FindWindow("Shell_TrayWnd", "");
            ShowWindow(hwnd, SW_SHOW);
            taskbarIsShown = true;
        }
        #endregion

        #region Keyboard and Mouse Hooks

        [DllImport("user32.dll", CharSet = CharSet.Auto, SetLastError = true)]
        private static extern IntPtr SetWindowsHookEx(int idHook, LowLevelKeyboardMouseProc lpfn, IntPtr hMod, uint dwThreadId);

        [DllImport("user32.dll", CharSet = CharSet.Auto, SetLastError = true)]
        [return: MarshalAs(UnmanagedType.Bool)]
        private static extern bool UnhookWindowsHookEx(IntPtr hhk);

        [DllImport("user32.dll", CharSet = CharSet.Auto, SetLastError = true)]
        private static extern IntPtr CallNextHookEx(IntPtr hhk, int nCode, IntPtr wParam, IntPtr lParam);

        [DllImport("kernel32.dll", CharSet = CharSet.Auto, SetLastError = true)]
        private static extern IntPtr GetModuleHandle(string lpModuleName);

        [DllImport("user32.dll", CharSet = CharSet.Auto, ExactSpelling = true, CallingConvention = CallingConvention.Winapi)]
        public static extern short GetKeyState(int keyCode);

        private const int WH_KEYBOARD_LL = 13;
        private const int WH_MOUSE_LL = 14;
        private const int WM_KEYDOWN = 0x0100;
        private const int WM_KEYUP = 0x0101;
        private const int WM_SYSKEYDOWN = 0x0100;
        private const int WM_SYSKEYUP = 0x0101;
        private static LowLevelKeyboardMouseProc keyboardProc = KeyboardHookCallback;
        private static LowLevelKeyboardMouseProc mouseProc = MouseHookCallback;
        private static IntPtr _hookID1 = IntPtr.Zero;
        private static IntPtr _hookID2 = IntPtr.Zero;

        private delegate IntPtr LowLevelKeyboardMouseProc(int nCode, IntPtr wParam, IntPtr lParam);

        private static IntPtr SetHookKeyboard(LowLevelKeyboardMouseProc proc)
        {
            using (Process curProcess = Process.GetCurrentProcess())
            using (ProcessModule curModule = curProcess.MainModule)
            {
                return SetWindowsHookEx(WH_KEYBOARD_LL, proc, GetModuleHandle(curModule.ModuleName), 0);
            }
        }

        private static IntPtr SetHookMouse(LowLevelKeyboardMouseProc proc)
        {
            using (Process curProcess = Process.GetCurrentProcess())
            using (ProcessModule curModule = curProcess.MainModule)
            {
                return SetWindowsHookEx(WH_MOUSE_LL, proc, GetModuleHandle(curModule.ModuleName), 0);
            }
        }

        private enum MouseMessages
        {
            WM_LBUTTONDOWN = 0x0201,
            WM_LBUTTONUP = 0x0202,
            WM_MOUSEMOVE = 0x0200,
            WM_MOUSEWHEEL = 0x020A,
            WM_RBUTTONDOWN = 0x0204,
            WM_RBUTTONUP = 0x0205,
            WM_MBUTTONDOWN = 0x0207,
            WM_MBUTTONUP = 0x0208,
        }

        [StructLayout(LayoutKind.Sequential)]
        private struct POINT
        {
            public int x;
            public int y;
        }

        [StructLayout(LayoutKind.Sequential)]
        private struct MSLLHOOKSTRUCT
        {
            public POINT pt;
            public uint mouseData;
            public uint flags;
            public uint time;
            public IntPtr dwExtraInfo;
        }

        [DllImport("user32.dll")]
        static extern bool SetCursorPos(int X, int Y);

        internal static ushort HIWORD(IntPtr dwValue)
        {
            return (ushort)((((long)dwValue) >> 0x10) & 0xffff);
        }

        internal static ushort HIWORD(uint dwValue)
        {
            return (ushort)(dwValue >> 0x10);
        }

        internal static int GET_WHEEL_DELTA_WPARAM(IntPtr wParam)
        {
            return (short)HIWORD(wParam);
        }

        internal static int GET_WHEEL_DELTA_WPARAM(uint wParam)
        {
            return (short)HIWORD(wParam);
        }

        #endregion

        /// <summary>
        /// The main entry point for the application.
        /// </summary>
        [STAThread]
        static void Main()
        {
            Application.EnableVisualStyles();
            Application.SetCompatibleTextRenderingDefault(false);

            frmMain f = new frmMain();
            f.SetStatuses(false, new bool[] { false, false, false, false });

            _hookID1 = SetHookKeyboard(keyboardProc);
            _hookID2 = SetHookMouse(mouseProc);
            NehebkauDevice.StartNotifier();

            Application.Run(f);

            UnhookWindowsHookEx(_hookID2);
        }

        private static byte KeyModifier;
        private static byte[] KeycodeBuffer = new byte[6];
        private static bool[] KeydownBuffer = new bool[256];

        private static int GetHid(int vkCode)
        {
            // alphabet letters
            if (vkCode >= 0x41 && vkCode <= 0x5A)
            {
                return vkCode - 0x41 + 0x04;
            }

            // top row numbers
            if (vkCode >= 0x31 && vkCode <= 0x39)
            {
                return vkCode - 0x31 + 0x1E;
            }

            if (vkCode == 0x30) return 0x27; // 0

            // numpad numbers
            if (vkCode >= 0x61 && vkCode <= 0x69)
            {
                return vkCode - 0x61 + 0x59;
            }

            if (vkCode == 0x60) return 0x62; // numpad 0

            // F1-F12
            if (vkCode >= 0x70 && vkCode <= 0x7B)
            {
                return vkCode - 0x70 + 0x3A;
            }

            // F13-F24
            if (vkCode >= 0x7C && vkCode <= 0x87)
            {
                return vkCode - 0x70 + 0x68;
            }

            switch (vkCode)
            {
                // OEM
                case 0xDB: return 0x2F;
                case 0xDC: return 0x31;
                case 0xDD: return 0x30;
                case 0xDE: return 0x34;
                case 0xBF: return 0x38;
                case 0xC0: return 0x35;
                case 0xBA: return 0x33;
                case 0xBB: return 0x2E;
                case 0xBC: return 0x36;
                case 0xBD: return 0x2D;
                case 0xBE: return 0x37;

                // locks
                case 0x90: return 0x53;
                case 0x91: return 0x47;
                case 0x14: return 0x39;
                
                // arrows
                case 0x25: return 0x50;
                case 0x26: return 0x52;
                case 0x27: return 0x4F;
                case 0x28: return 0x51;


                case 0x08: return 0x2A;
                case 0x09: return 0x2B;
                case 0x0C: return 0xA2;
                case 0x0D: return 0x28;
                case 0x10: return 0xE1;
                case 0x11: return 0xE0;
                case 0x12: return 0xE2;
                case 0x13: return 0x48;
                case 0x1B: return 0x29;
                case 0x20: return 0x2C;
                case 0x21: return 0x4B;
                case 0x22: return 0x4E;
                case 0x23: return 0x4D;
                case 0x24: return 0x4A;
                case 0x2C: return 0x46;
                case 0x2D: return 0x49;
                case 0x2E: return 0x4C;
                case 0x5B: return 0xE3;
                case 0x5C: return 0xE7;
                case 0x5D: return 0x65;
                case 0x5F: return 0x66; // sleep->power
                case 0x6A: return 0x55;
                case 0x6B: return 0x57;
                case 0x6D: return 0x56;
                case 0x6E: return 0x63;
                case 0x6F: return 0x54;
                case 0xA0: return 0xE1;
                case 0xA1: return 0xE5;
                case 0xA2: return 0xE0;
                case 0xA3: return 0xE4;
                case 0xA4: return 0xE3;
                case 0xA5: return 0xE7;
                case 0x2F: return 0x75;
                case 0xAD: return 0x7F;
                case 0xAE: return 0x81;
                case 0xAF: return 0x80;
            }
            return 0;
        }

        private static IntPtr KeyboardHookCallback(int nCode, IntPtr wParam, IntPtr lParam)
        {
            if (nCode >= 0)
            {
                int vkCode = Marshal.ReadInt32(lParam);
                int hidCode = GetHid(vkCode);

                if (wParam == (IntPtr)WM_KEYDOWN || wParam == (IntPtr)WM_SYSKEYDOWN)
                {
                    KeydownBuffer[hidCode] = true;
                }
                else if (wParam == (IntPtr)WM_KEYUP || wParam == (IntPtr)WM_SYSKEYUP)
                {
                    KeydownBuffer[hidCode] = false;
                }

                KeyModifier = 0;
                for (int i = 0; i < 8; i++)
                {
                    if (KeydownBuffer[0xE0 + i])
                        KeyModifier |= (byte)(1 << i);
                }

                for (int i = 0; i < KeycodeBuffer.Length; i++)
                {
                    KeycodeBuffer[i] = 0;
                }

                for (int i = 1, j = 0; i < KeydownBuffer.Length && j < KeycodeBuffer.Length; i++)
                {
                    if (KeydownBuffer[i])
                    {
                        KeycodeBuffer[j] = (byte)i;
                        j++;
                    }
                }
                
                if (NehebkauDevice.WriteKeyboard(KeyModifier, KeycodeBuffer) > 0)
                {
                    frmMain.OnlyInstance.AddToActivity();
                }
            }

            return CallNextHookEx(_hookID1, nCode, wParam, lParam);
        }

        private static int MouseButtonFlags = 0x00;
        public static int MouseLastX = 0;
        public static int MouseLastY = 0;
        public static bool allowMouseMove = true;

        private static IntPtr MouseHookCallback(int nCode, IntPtr wParam, IntPtr lParam)
        {
            if (nCode >= 0)
            {
                MSLLHOOKSTRUCT hookStruct = (MSLLHOOKSTRUCT)Marshal.PtrToStructure(lParam, typeof(MSLLHOOKSTRUCT));
                int dX = 0, dY = 0, dW = 0;
                switch ((MouseMessages)wParam)
                {
                    case MouseMessages.WM_MOUSEMOVE:
                        int curX = hookStruct.pt.x;
                        int curY = hookStruct.pt.y;
                        if (allowMouseMove)
                        {
                            dX = curX - MouseLastX;
                            dY = curY - MouseLastY;
                        }
                        MouseLastY = curY;
                        MouseLastX = curX;
                        break;
                    case MouseMessages.WM_MOUSEWHEEL:
                        dW = GET_WHEEL_DELTA_WPARAM(hookStruct.mouseData);
                        break;
                    case MouseMessages.WM_LBUTTONDOWN:
                        MouseButtonFlags |= (1 << 0);
                        break;
                    case MouseMessages.WM_LBUTTONUP:
                        MouseButtonFlags &= ~(1 << 0);
                        break;
                    case MouseMessages.WM_RBUTTONDOWN:
                        MouseButtonFlags |= (1 << 2);
                        break;
                    case MouseMessages.WM_RBUTTONUP:
                        MouseButtonFlags &= ~(1 << 2);
                        break;
                    case MouseMessages.WM_MBUTTONDOWN:
                        MouseButtonFlags |= (1 << 1);
                        break;
                    case MouseMessages.WM_MBUTTONUP:
                        MouseButtonFlags &= ~(1 << 1);
                        break;
                }

                if (NehebkauDevice.WriteMouse(MouseButtonFlags, dX, dY, dW) > 0)
                {
                    frmMain.OnlyInstance.AddToActivity();
                }
            }

            return CallNextHookEx(_hookID2, nCode, wParam, lParam);
        }
    }
}
